! ***********************************************************************
!
!   Copyright (C) 2010  Bill Paxton
!
!   this file is part of mesa.
!
!   mesa is free software; you can redistribute it and/or modify
!   it under the terms of the gnu general library public license as published
!   by the free software foundation; either version 2 of the license, or
!   (at your option) any later version.
!
!   mesa is distributed in the hope that it will be useful, 
!   but without any warranty; without even the implied warranty of
!   merchantability or fitness for a particular purpose.  see the
!   gnu library general public license for more details.
!
!   you should have received a copy of the gnu library general public license
!   along with this software; if not, write to the free software
!   foundation, inc., 59 temple place, suite 330, boston, ma 02111-1307 usa
!
! ***********************************************************************
 
      module run_star_extras

      use star_lib
      use star_def
      use const_def
      
      implicit none
      
      integer :: time0, time1, clock_rate
      double precision, parameter :: expected_runtime = 1.6 ! minutes
      
      ! these routines are called by the standard run_star check_model
      contains
      
      subroutine extras_controls(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine extras_controls
      
!MANOS ADDS - NOTES ARE FROM other_am_mixing.f and other_mixing.f from /mesa/star/other

      ! NOTE: remember to set use_other_am_mixing = .true. to enable this.
      
      ! you can add your own mixing routine for use in addition to the standard ones.
      ! don't edit this file -- instead copy null_other_am_mixing from below and edit it.
      ! it goes in your src/run_star_extras file.
      ! edit the extras_controls routine to set s% other_am_mixing
      ! e.g.,
         !    s% other_am_mixing => my_am_mixing_routine
      
      ! you can revise vector s% am_nu(:) of diffusion coefficients for angular momentum
      ! when your routine is called, am_nu has been set to standard values.
      ! you can change these as desired. 
      
         ! for example, you can add uniform extra diffusion as follows: 
         
            ! nu_additional = 3d4 ! cm^2/s
            ! forall (k=1:s% nz) s% am_nu(k) = s% am_nu(k) + nu_additional

         
         
      ! NOTE: if you'd like to have some inlist controls for your routine,
      ! you can use the x_ctrl array of real(dp) variables that is in &controls
      ! e.g., in the &controls inlist, you can set
      !     x_ctrl(1) = my_special_param
      ! then in your routine, you can access that by
      !     s% x_ctrl(1)
      ! of course before you can use s, you need to get it using the id argument.
      ! here's an example of how to do that -- add these lines at the start of your routine:
      !         use star_lib, only: star_ptr
      !         type (star_info), pointer :: s
      !         call star_ptr(id, s, ierr)
      !         if (ierr /= 0) then ! OOPS
      !            return
      !         end if
      ! 
      ! for integer control values, you can use x_integer_ctrl
      ! for logical control values, you can use x_logical_ctrl

      subroutine null_other_D_mix(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
	 type (star_info), pointer :: s
	 integer :: k
	 real(dp) :: D_mri,lnLambda,eta,kappa,ftr1,mri_crit,shear2
         ierr = 0
	 
	 
	 ftr1=1.
	 
	 call star_ptr(id, s, ierr)
	 if (ierr /=0) return
	 
	 lnLambda=0.
	 eta=0.
	 kappa=0.
	 mri_crit=0.
	 D_mri=0.
	 
	 shear2=0.
	 
	 do k=1,s% nz	 
	 
	 shear2=max(-1.e+30,s% domega_dlnr(k)/(s% omega(k)))
	 
	    lnLambda = log(2/(3*qe**3) &
            *sqrt(mp*(s% abar(k))*(kerg*(s% T(k)))**3/(pi*(s% rho(k))* &
            (max(1d-10,s% zbar(k)))**5)))
	    if (lnLambda < 1) lnLambda = 1

	    eta = 7.0d11*lnLambda/(s% T(k))**1.5d0
	    
	    kappa = 16*boltz_sigma*(s% T(k))**3/(3*(s% opacity(k))* &
	   (s% rho(k))**2*(s% Cp(k)))	 
	   
	    mri_crit = (eta/kappa)*(s% brunt_N2(k) - &
	    s% brunt_N2_composition_term(k))*ftr1 + &
	    s% brunt_N2_composition_term(k) &
	    + 2.*((s% omega(k))**2)*(shear2)

	   
	 if(mri_crit < 0) then
	 !WKC  	 
	 D_mri = 0.02*abs(shear2)*&
	 (s% omega(k))*(s% r(k))**2
	 !Spruit 99
	 !D_mri = 0.1*(abs(shear2))*kappa*(s% omega(k)/&
	 !(s% brunt_N2(k)))**2
	 !Spruit 14
	 !D_mri = kappa
	 !Spruit NEW
	 !if(s% brunt_N2(k) /= 0)then
	 !D_mri = abs(shear2)*(s% omega(k))*&
	 !min(s% r(k),s% scale_height(k))*(kappa/(abs(s% brunt_N2(k)))**0.5)**0.5
	 !else
	 !D_mri = 0.0
	 !endif
	 else
	 D_mri = 0.0
	 endif
	 	 
         s% D_mix(k) = s% D_mix(k) + D_mri !*0.03333d0
	 
	 enddo	 
	 
      end subroutine null_other_D_mix
      
      subroutine null_other_am_mixing(id, ierr)
         integer, intent(in) :: id
         integer, intent(out) :: ierr
	 type (star_info), pointer ::s
	 integer :: k
	 real(dp) :: nu_mri,lnLambda,eta,kappa,ftr1,mri_crit,shear2
         ierr = 0
	 
	 ftr1=1.
	 
	 call star_ptr(id, s, ierr)
	 if(ierr /=0) return
	 
	 lnLambda=0.
	 eta=0.
	 kappa=0.
	 mri_crit=0.
	 nu_mri=0.
	 
	 shear2=0.
	 	 
	 do k=1,s% nz
	 
	 shear2=max(-1.e+30,s% domega_dlnr(k)/(s% omega(k)))
	 
	    lnLambda = log(2/(3*qe**3) &
            *sqrt(mp*(s% abar(k))*(kerg*(s% T(k)))**3/(pi*(s% rho(k))* &
            (max(1d-10,s% zbar(k)))**5)))
	    if (lnLambda < 1) lnLambda = 1

	    eta = 7.0d11*lnLambda/(s% T(k))**1.5d0
	    
	    kappa = 16*boltz_sigma*(s% T(k))**3/(3*(s% opacity(k))* &
	   (s% rho(k))**2*(s% Cp(k)))	 
	   
	    mri_crit = (eta/kappa)*(s% brunt_N2(k) - &
	    s% brunt_N2_composition_term(k))*ftr1 + &
	    s% brunt_N2_composition_term(k) &
	    + 2.*((s% omega(k))**2)*(shear2)	 
	 
	 
	 if(mri_crit < 0)then	 
	 nu_mri = 0.02*abs(shear2)*&
	 (s% omega(k))*(s% r(k))**2
	 !Spruit 99
	 !nu_mri = 0.1*(abs(shear2))*kappa*(s% omega(k)/&
	 !(s% brunt_N2(k)))**2
	 !Spruit 14
	 !nu_mri = kappa
	 !Spruit NEW
	 !if(s% brunt_N2(k) /= 0)then
	 !nu_mri = abs(shear2)*(s% omega(k))*&
	 !min(s% r(k),s% scale_height(k))*(kappa/(abs(s% brunt_N2(k)))**0.5)**0.5
	 !else
	 !nu_mri = 0.0
	 !endif
	 else
	 nu_mri = 0.0
	 endif 
	 
         s% am_nu_rot(k) = s% am_nu_rot(k) + nu_mri
	 enddo
	 
      end subroutine null_other_am_mixing

!END OF MANOS ADDS~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
      
      
      integer function extras_startup(id, restart, ierr)
         integer, intent(in) :: id
         logical, intent(in) :: restart
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_startup = 0
         call system_clock(time0,clock_rate)
         if (.not. restart) then
            call alloc_extra_info(s)
         else ! it is a restart
            call unpack_extra_info(s)
         end if
      end function extras_startup

      subroutine extras_after_evolve(id, id_extra, ierr)
         use num_lib
         integer, intent(in) :: id, id_extra
         integer, intent(out) :: ierr
         double precision :: dt
         integer :: k, k_cntr, k_surf
         type (star_info), pointer :: s
         logical :: okay
         character (len=strlen) :: test
         
         include 'formats'

         
         okay = .true.
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         call system_clock(time1,clock_rate)
         dt = dble(time1 - time0) / clock_rate / 60
         call GET_ENVIRONMENT_VARIABLE( &
            "MESA_TEST_SUITE_CHECK_RUNTIME", test, status=ierr, trim_name=.true.)
         if (ierr == 0 .and. trim(test) == 'true' .and. dt > 1.5*expected_runtime) then
            write(*,'(/,a70,2f12.1,99i10/)') &
               'failed: EXCESSIVE runtime, prev time, retries, backups, steps', &
               dt, expected_runtime, s% num_retries, s% num_backups, s% model_number
         else
            write(*,'(/,a50,2f12.1,99i10/)') 'runtime, prev time, retries, backups, steps', &
               dt, expected_runtime, s% num_retries, s% num_backups, s% model_number
         end if
         ierr = 0
         
         write(*,*)
         call check('star_mass', s% star_mass, 12.5d0, 14.0d0)
         call check('log total_angular_momentum', safe_log10_cr(s% total_angular_momentum), 50d0, 52.5d0)
         call check('log center_omega', safe_log10_cr(s% center_omega), -5.7d0, -4.2d0)
         call check('log he_core_omega', safe_log10_cr(s% he_core_omega), -5.7d0, -4.2d0)
         call check('he_core_mass', s% he_core_mass, 4.0d0, 5.1d0)
         call check('surface j_rot', safe_log10_cr(s% j_rot(1)),  17d0, 19d0)
         call check('surface v_rot', s% omega(1)*s% r(1)*1d-5, 0.25d0, 0.45d0)
         
         k_cntr = 0
         k_surf = 0
         do k = s% nz, 1, -1
            if (s% m(k) > 3.0*Msun .and. k_cntr == 0) k_cntr = k
            if (s% m(k) > 3.4*Msun .and. k_surf == 0) k_surf = k
         end do
         
         write(*,*)
         write(*,1) 'avg from 3.0 to 3.4 Msun'
         call check('logT', avg_val(s% lnT)/ln10, 7.5d0, 8.1d0)
         call check('logRho', avg_val(s% lnd)/ln10, 1d0, 3.0d0)
         call check('log j_rot', safe_log10_cr(avg_val(s% j_rot)), 15.2d0, 16.4d0)
         call check('D_ES', safe_log10_cr(avg_val(s% D_ES)), 2.1d0, 4.9d0) 
         call check('D_ST', safe_log10_cr(avg_val(s% D_ST)), 2d0, 8d0)
         call check('nu_ST', safe_log10_cr(avg_val(s% nu_ST)), 8.1d0, 10.9d0)
         call check('dynamo_B_r', safe_log10_cr(avg_val(s% dynamo_B_r)), 0.2d0, 1d0)
         call check('dynamo_B_phi', safe_log10_cr(avg_val(s% dynamo_B_phi)), 3d0, 7.5d0)
         write(*,*)
         if (okay) write(*,'(a)') 'all values are within tolerances'
         write(*,*)
         
         
         contains
         
         double precision function avg_val(v)
            double precision :: v(:)
            avg_val = dot_product(v(k_surf:k_cntr), s% dq(k_surf:k_cntr)) / sum(s% dq(k_surf:k_cntr))
         end function avg_val
         
         subroutine check(str, val, low, hi)
            double precision, intent(in) :: val, low, hi
            character (len=*) :: str
            include 'formats'
            if (low <= val .and. val <= hi) then
               write(*,1) trim(str), val, low, hi
            else
               write(*,1) '*** BAD *** ' // trim(str), val, low, hi
               okay = .false.
            end if
         end subroutine check
         
         
      end subroutine extras_after_evolve      

      ! returns either keep_going, retry, backup, or terminate.
      integer function extras_check_model(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_check_model = keep_going         
      end function extras_check_model

      integer function how_many_extra_history_columns(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_history_columns = 0
      end function how_many_extra_history_columns
      
      
      subroutine data_for_extra_history_columns(id, id_extra, n, names, vals, ierr)
         integer, intent(in) :: id, id_extra, n
         character (len=maxlen_history_column_name) :: names(n)
         real(dp) :: vals(n)
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
      end subroutine data_for_extra_history_columns

      
      integer function how_many_extra_profile_columns(id, id_extra)
         use star_def, only: star_info
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         how_many_extra_profile_columns = 13
      end function how_many_extra_profile_columns
      
      
      subroutine data_for_extra_profile_columns(id, id_extra, n, nz, names, vals, ierr)
         use star_def, only: star_info, maxlen_profile_column_name
         use const_def, only: dp
         integer, intent(in) :: id, id_extra, n, nz
         character (len=maxlen_profile_column_name) :: names(n)
         real(dp) :: vals(nz,n),lnLambda,eta,kappa,ftr1,alfa,beta,delta(nz),shear2
         integer, intent(out) :: ierr
         type (star_info), pointer :: s
         integer :: k
         ierr = 0
         call star_ptr(id, s, ierr)
	 
	 ftr1=1.
	 
	 !if (n /= 1) stop 'data_for_extra_profile_columns'
	 names(1) = 'eta_over_kappa'
	 names(2) = 'mri_criterion'
	 names(3) = 'D_mri'
	 names(4) = 'my_shear'
	 names(5) = 'S_wkc'
	 names(6) = 'D_s99'
	 names(7) = 'S_s99'
	 names(8) = 'D_s14'
	 names(9) = 'S_s14'
	 names(10) = 'eta'
	 names(11) = 'xi'
	 names(12) = 'D_newS'
	 names(13) = 'S_newS'
	 
	 
	 shear2=0.
	 	 
	 do k = 1,nz
	 
	 !shear2=max(1d-30,min(1d30,abs(s% domega_dlnr(k)/s% omega(k))))
	 shear2=max(-1.e+30,s% domega_dlnr(k)/(s% omega(k)))
	 vals(k,4) = shear2

	 
	    lnLambda = log(2/(3*qe**3) &
            *sqrt(mp*(s% abar(k))*(kerg*(s% T(k)))**3/(pi*(s% rho(k))* &
            (max(1d-10,s% zbar(k)))**5)))
	    if (lnLambda < 1) lnLambda = 1

	    eta = 7.0d11*lnLambda/(s% T(k))**1.5d0
	    
	    kappa = 16*boltz_sigma*(s% T(k))**3/(3*(s% opacity(k))* &
	   (s% rho(k))**2*(s% Cp(k)))
	    
	    vals(k,1) = eta/kappa
	    vals(k,10) = eta
	    vals(k,11) = kappa
	    
	    vals(k,2) = vals(k,1)*(s% brunt_N2(k) - &
	    s% brunt_N2_composition_term(k))*ftr1 + &
	    s% brunt_N2_composition_term(k) &
	    + 2.*((s% omega(k))**2)*(shear2)
	    
	    if(vals(k,2) < 0)then !FIX LATER!<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	    !WKC 14 nu and S    
	    vals(k,3) = 0.02*abs(shear2)*&
	    (s% omega(k))*(s% r(k))**2   
	    vals(k,5) = vals(k,3)*(s% rho(k))*abs(shear2)*(s% omega(k))
	    !Spruit 99
	    vals(k,6) = 0.1*abs(shear2)*kappa*(s% omega(k) &
	    /(s% brunt_N2(k)))**2
	    vals(k,7) = vals(k,5)*(s% rho(k))*abs(shear2)*(s% omega(k))
	    !Spruit 14
	    vals(k,8) = kappa
	    vals(k,9) = vals(k,8)*(s% rho(k))*abs(shear2)*(s% omega(k))
	    !Spruit NEW
	    if(s% brunt_N2(k) /= 0)then
	    vals(k,12) = abs(shear2)*(s% omega(k))*&
	    min(s% r(k),s% scale_height(k))*(kappa/(abs(s% brunt_N2(k)))**0.5)**0.5
	    vals(k,13) = vals(k,12)*(s% rho(k))*abs(shear2)*(s% omega(k))
	    else
	    vals(k,12) = 0.0
	    vals(k,13) = 0.0
	    endif
	    else	    
	    vals(k,3) = 0.0
	    vals(k,5) = 0.0
	    vals(k,6) = 0.0
	    vals(k,7) = 0.0
	    vals(k,8) = 0.0
	    vals(k,9) = 0.0
	    vals(k,12) = 0.0
	    vals(k,13) = 0.0
	    endif
	 
	 enddo
	 
	 !vals(1,1) = 0.
         
         !note: do NOT add these names to profile_columns.list
         ! the profile_columns.list is only for the built-in profile column options.
         ! it must not include the new column names you are adding here.

         ! here is an example for adding a profile column
         !if (n /= 1) stop 'data_for_extra_profile_columns'
         !names(1) = 'beta'
         !do k = 1, nz
         !   vals(k,1) = s% Pgas(k)/s% P(k)
         !end do
          if (ierr /= 0) return   
      end subroutine data_for_extra_profile_columns
      

      ! returns either keep_going or terminate.
      ! note: cannot request retry or backup; extras_check_model can do that.
      integer function extras_finish_step(id, id_extra)
         integer, intent(in) :: id, id_extra
         integer :: ierr
         type (star_info), pointer :: s
         ierr = 0
         call star_ptr(id, s, ierr)
         if (ierr /= 0) return
         extras_finish_step = keep_going
         call store_extra_info(s)
      end function extras_finish_step
      
      
      ! routines for saving and restoring extra data so can do restarts
         
         ! put these defs at the top and delete from the following routines
         !integer, parameter :: extra_info_alloc = 1
         !integer, parameter :: extra_info_get = 2
         !integer, parameter :: extra_info_put = 3
      
      
      subroutine alloc_extra_info(s)
         integer, parameter :: extra_info_alloc = 1
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_alloc)
      end subroutine alloc_extra_info
      
      
      subroutine unpack_extra_info(s)
         integer, parameter :: extra_info_get = 2
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_get)
      end subroutine unpack_extra_info
      
      
      subroutine store_extra_info(s)
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         call move_extra_info(s,extra_info_put)
      end subroutine store_extra_info
      
      
      subroutine move_extra_info(s,op)
         integer, parameter :: extra_info_alloc = 1
         integer, parameter :: extra_info_get = 2
         integer, parameter :: extra_info_put = 3
         type (star_info), pointer :: s
         integer, intent(in) :: op
         
         integer :: i, j, num_ints, num_dbls, ierr
         
         i = 0
         ! call move_int or move_flg    
         num_ints = i
         
         i = 0
         ! call move_dbl       
         
         num_dbls = i
         
         if (op /= extra_info_alloc) return
         if (num_ints == 0 .and. num_dbls == 0) return
         
         ierr = 0
         call star_alloc_extras(s% id, num_ints, num_dbls, ierr)
         if (ierr /= 0) then
            write(*,*) 'failed in star_alloc_extras'
            write(*,*) 'alloc_extras num_ints', num_ints
            write(*,*) 'alloc_extras num_dbls', num_dbls
            stop 1
         end if
         
         contains
         
         subroutine move_dbl(dbl)
            double precision :: dbl
            i = i+1
            select case (op)
            case (extra_info_get)
               dbl = s% extra_work(i)
            case (extra_info_put)
               s% extra_work(i) = dbl
            end select
         end subroutine move_dbl
         
         subroutine move_int(int)
            integer :: int
            i = i+1
            select case (op)
            case (extra_info_get)
               int = s% extra_iwork(i)
            case (extra_info_put)
               s% extra_iwork(i) = int
            end select
         end subroutine move_int
         
         subroutine move_flg(flg)
            logical :: flg
            i = i+1
            select case (op)
            case (extra_info_get)
               flg = (s% extra_iwork(i) /= 0)
            case (extra_info_put)
               if (flg) then
                  s% extra_iwork(i) = 1
               else
                  s% extra_iwork(i) = 0
               end if
            end select
         end subroutine move_flg
      
      end subroutine move_extra_info

      end module run_star_extras
